#' Coerces input to image (array)
#' @inheritParams base::as.array
#' @param ... see \code{\link[base]{as.array}}
#' @family convert
#' @export
as.image <- as.array

#' Coerces input to image, shorthand for as.image
#' @param arr array to convert
#' @param transpose boolean, whether to transpose the resultant array.
#'
#'                 Note that Voir images are of dimension (nx,ny,nchanns),
#'                 and the returned array will also have dimensions
#'                 (nx,ny,nchanns).
#'              
#'                 That is, *the returned array will be transposed*, since
#'                  usually an image as an array would have dimensions
#'                  (ny,nx,nchanns), being (rows,columns,channels).
#'         
#'                 If \code{transpose=TRUE}, then the returned array will
#'                 have dimensions (ny, nx, nchanns).
#' @param ... IGNORED
#' @family convert
#' @export
array2im <- function(arr,transpose=FALSE,...)
{
    if ( transpose )
        arr <- imtranspose(arr)
    return( as.image(arr) )
}

#' Coerces image to array.
#' @inheritParams array2im
#' @param img image to convert
#' @family convert
#' @export
im2array <- function(img,transpose=FALSE,...)
{ 
    if (transpose)
        img <- imtranspose(img)
    return(as.array(img))
}

#' Checks whether the input is an image
#' @family convert
#' @export
is.image <- is.array


# #  NOTE: I like these versions better but if the image is 0 to 1 and you
# #  round it, you lose information.
# #  if you convert 0 to 255 then you can have unintended consequences when
# #  you further use the images.
# # ' Convert image to char (1-byte uint)
# # ' @param img image to convert
# # ' @return image as char (0-255).
# # ' @family convert
# # ' @export
# imchar <- function(img)
# {
#     return(roundAndCap(img,0,255))
# }
# 
# # ' Convert image to 4-byte int (-2^31+1 to 2^31)
# # ' @inheritParams imchar
# # ' @return image as char (-2^31+1 to 2^31)
# # ' @family convert
# # ' @export
# imint <- function(img)
# {
#     return(roundAndCap(img,-2^31+1,2^31))
# }

#' Convert image to char (DOESN'T DO ANYTHING)
#' @param img image to convert
#' @return input image.
#' 
#' @note These functions just return the input image as-is. 
#' They *could* scale the images to the appropriate range (0 to 1 maps to 0
#' to 255, e.g.) and round, but then one has to remember that the images no
#' longer range from 0 to 1 (which is the default for all the other
#' functions in this package).
#' @aliases imint
#' @rdname imchar
#' @family convert
#' @seealso \code{\link{imbinary}}, which *does* do converting.
#' @export
imchar <- function(img)
{
    return(img)
}
#' Convert image to 4-byte int (-2^31+1 to 2^31) (DOESN'T DO ANYTHING)
#' @inheritParams imchar
#' @family convert
#' @rdname imchar
#' @export
imint <- function(img)
{
    return(img)
}

#' Convert image to logical
#' @inheritParams imchar
#' @return image as binary (TRUE/FALSE)
#' @family convert
#' @export
imbinary <- function(img)
{
    return(as.logical(img))
}

#' converts to int and caps to range.
#' @keywords internal
#' @export
roundAndCap <- function(x, MIN, MAX, max=imageUtilities.option('max'))
{
    return(pmin(pmax(round(x),MIN),MAX))
}


#' gets number of pixels in the horizontal (x) direction
#' @param img image to get dimensions of
#' @param ... not used (backwards compatibility)
#' @return number of x pixels
#' @family dims
#' @export
imgetnx <- function(img,...)
{
    return(dim(img)[2])
}

#' gets number of pixels in the vertical (y) direction
#' @inheritParams imgetnx
#' @return number of y pixels
#' @family dims
#' @export
imgetny <- function(img,...)
{
    return(dim(img)[1])
}

#' gets dimensions of image, x, y, channs
#' @inheritParams imgetnx
#' @return vector (nx, ny, numchans)
#' @family dims
#' @export
imgetdims <- function(img)
{
    return( c( ncol(img), nrow(img), imgetnumcomp(img) ) )
}

#' gets number of image channels/components
#' @inheritParams imgetnx
#' @return number of channels.
#' @family dims
#' @export
imgetnumcomp <- function(img)
{
    return( ifelse(length(dim(img))==2,1,dim(img)[3]) )
}

#' downsamples an image
#' @param imgA   input image to be downsampled
#' @param dx     window size (horizontal dimension)
#' @param dy     window size (vertical dimension), default dx.
#' @param method one of "every" or "mean".
#'               If "every", every dx and dy'th pixel will be kept
#'                (starting from the first)
#'               If "mean", the mean of every non-overlapping dx by dy box
#'                will be calculated and that is retained.
#' @param border.fill if the image dimensions are not exactly divisible by
#' dx or dy, then what value will be used in the border (right/bottom)?
#' Default 0, but 'NA' is another option. Only relevant for 'mean' method.
#' @return downsampled image, of size ceiling(dim(imgA)/c(dy,dx)), where
#'         pixel (i,j) is calculated according to parameter "method".
#' @note \code{imsubsample} and \code{imdownsample} are aliases for the
#' same function.
#' @details 
#'      The \code{method="mean"} mode is \bold{very slow} at the moment. It
#'       takes ~11 user-seconds on a 5000x1000x3 image with dx=3, and
#'       ~28 user-seconds on a 5000x3000x3 image, dx=3.
#'      Consider whether you actually \emph{need} the mean in each cell
#'       before using this method; for display purposes
#'       \code{method="every"} does fine.
#'  
#'      \bold{UPDATE (v1.1)}: \code{method="mean"} has been sped up and
#'       now takes about 6 user-seconds on 5000x3000x3, dx=3 (but still a
#'       decent amount of time overall).
#'
#'       It is **much much** faster if \code{dx} and \code{dy} 
#'       exactly divide the dimensions of the input array, or else it 
#'       makes a copy of the input array which slows it down (TODO: remove
#'       this bottle-neck).
#'
#' @aliases imsubsample imdownsample
#' @rdname imdownsample
#' @family subsample
#' @export
imdownsample <- function( imgA, dx, dy=dx, method=c('every','mean'),
                         border.fill=0 )
{
    method <- match.arg(method)
    imgA <- imexpand(imgA)
    if ( method == "every" ) {
        return( imsqueeze(imgA[ seq(1,nrow(imgA),by=dy), 
                                seq(1,ncol(imgA),by=dx), ] ))
    } else if ( method == "mean" ) {
        # NEW METHOD!
        sz <- dim(imgA)
        inSz <- floor(sz[1:2]/c(dy,dx))*c(dy,dx)

        if ( any(sz[1:2] != inSz) ) {
            # ARGH: The extra allocating KILLS.
            # Could speed it up by doing the subsample of the smallest
            # fitting array, and then gluing on the other bits?
            # In fact, this is the bottle-neck.
            
            outSz <- c(ceiling(sz[1:2]/c(dy,dx)),sz[3])
            imgO <- array(border.fill,dim=outSz*c(dy,dx,1))
            # SLOW SLOW SLOW
            imgO[1:sz[1],1:sz[2],] <- imgA
        } else {
            imgO <- imgA
        }
        return( imsqueeze(subsample.rows(subsample.cols(imgO,
                                                        dx,na.rm=TRUE),
                                         dy,na.rm=TRUE)) )
#        sz <- dim(imgA)
#        outSz <- c(ceiling(sz[1:2]/c(dy,dx)),sz[3])
#        is <- pmin(seq(1,sz[1],by=dy) + floor(dy/2),sz[1]-1)
#        js <- pmin(seq(1,sz[2],by=dx) + floor(dx/2),sz[2]-1)
#        out3 <- array(NA,dim=outSz)
#        for ( c in 1:sz[3] ) {
#            out3[,,c] <- (t(filter(
#                            t(filter(imgA[,,c],rep(1,dy))),
#                            rep(1,dx))))[is,js]
#        }
#        out3 <- out3/(dx*dy)
#        return(imsqueeze(out3))
    }
}

#' subsample an array, averaging every dy rows.
#' @param x array to subsample
#' @param dy how many rows to average over
#' @param ... passed into \code{colMeans}
#' @return array of the same size as \code{x}, except number of rows is
#' \code{nrow(x)/dy}.
#' @note assumes \code{nrow(x)} is exactly divisible by \code{dy}
#' @export
#' @family subsample
#' @keywords internal
subsample.rows <- function(x,dy,...)
{
    if ( dy == 1 ) return(x)
    sz <- dim(x)
    if ( sz[1] %% dy != 0 )
        stop('Number of rows/cols not divisible by dy/dx')
    dim(x) <- c(dy,prod(sz)/dy)
    sz[1] <- sz[1]/dy
    return(array( colMeans(x,...), dim=sz ))
}

#' subsample an array, averaging every dx cols.
#' @param x array to subsample
#' @param dx how many cols to average over
#' @param ... passed into \code{rowMeans}
#' @note assumes \code{nrow(x)} is exactly divisible by \code{dy}
#' @return array of the same size as \code{x}, except number of cols is
#' \code{ncol(x)/dx}.
#' @export
#' @family subsample
#' @keywords internal
subsample.cols <- function(x,dx,...)
{
    if ( dx == 1 ) return(x)
    return(imtranspose(subsample.rows(imtranspose(x),dx,...)))
}

#' @rdname imdownsample
#' @export
imsubsample <- imdownsample

#' combines single-channel images together to become multi-channel
#' @param ... input images to join, all of the same dimensions.
#' @return an image in which each of the input arguments forms a component
#' @export
imjoin <- function(...)
{
    ims <- list(...)
    if ( length(ims) == 1 )
        return(ims[[1]])

    if ( !all(sapply(ims,is.image)) )
        stop("imjoin: all inputs must be images!")
    # drop singletons.
    ims <- lapply(ims,drop)
    sz <- dim(ims[[1]])
    if ( length(sz) > 2 )
        stop("imjoin: all inputs must be single-channel!")
    if ( is.list((dims <- sapply(ims,dim))) ||
         !all(sz==dims) )
        stop("imjoin: all inputs must have same dimensions!")

    outI <- array( unlist(ims),dim=c(sz,length(ims)) )
}

#' Produces a subset of an image
#' @param img input image to subset
#' @param fx  first x/col (1-based)
#' @param fy  first y/row (1-based)
#' @param lx  last x/col (1-based)
#' @param ly  last y/row (1-based)
#' @param one.based boolean, default TRUE, whether fx, fy, lx and ly are
#'                  one or zero based.
#' @return a (ly-fy+1) by (lx-fx+1) by (numchanns) array containing the
#' subset of the image from fx to lx, fy to ly inclusive.
#' @note This should mask Voir's imsubset, however, it may not.
#' @export
imsubset <- function( img, fx, fy, lx, ly, one.based=TRUE )
{
    if ( !one.based ) {
        fx <- fx+1
        fy <- fy+1
        lx <- lx+1
        ly <- ly+1
    }
    if ( length(dim(img))==2 )
        return( img[fy:ly,fx:lx] )
    else
        return( img[fy:ly,fx:lx,] )
}

#' pastes patch image into src image starting at (src) fx,fy
#' @param src source image
#' @param patch image to paste into src, of same number of channels as src.
#' @param fx start x/row of src to paste patch into.
#' @param fy start y/col of src to paste patch into.
#' @param one.based whether fx and fy are one-based or 0-based.
#' @return the image src with patch pasted in at fx, fy
#' @export
impaste <- function( patch, src, fx, fy, one.based=TRUE ) 
{
    patch <- imexpand(patch)
    srcI  <- imexpand(src)

    if ( dim(patch)[3] != dim(srcI)[3] )
        stop("impaste: src and patch images should have same number of channels!")

    if ( !one.based ) { 
        fx <- fx + 1 
        fy <- fy + 1 
    } 

    sz   <- dim(patch)
    srcI[ fy:(fy+sz[1]-1), fx:(fx+sz[2]-1), ] <- patch

    return(imsqueeze(srcI))
}

