# NOT EXPORTED YET: 3D input images? transpose? etc.
#' 3x3 sobel operator
#' @param img image to apply operator to
#' @param which leave as is ('both') to calculate the traditional sobel
#' gradient. Use 'horizontal' or 'vertical' to calculate *just* those
#' gradients.
#' @return sobel gradient image: same size as input (the border will be NA
#' because of the convolution).
#' @details
#' The horizontal sobel gradient is:
#' \ifelse{latex}{
#' \deqn{
#' G_x = \left( \begin{array}{ccc}
#'  -1 & 0 & +1 \\
#'  -2 & 0 & +2 \\
#'  -1 & 0 & +1 
#' \end{array} \right) * img
#' }
#' }{
# Argh problems with roxygen doing word wrapping so I can't do ascii
# equations here.
#' \deqn{G_x = [-1 0 +1; -2 0 +2; -1 0 +1] * img}
#' }
#'
#' The vertical is:
#' \ifelse{latex}{
#' \deqn{
#' G_y = \left( \begin{array}{ccc}
#'  -1 & -2 & -1 \\
#'  0 & 0 & 0 \\
#'  +1 & +2 & +1 
#' \end{array} \right) * img
#' }
#' }{
# Argh problems with roxygen doing word wrapping so I can't do ascii
# equations here.
#' \deqn{G_y = [-1 -2 -1; 0 0 0; +1 +2 +1] * img}
#' }
#' where '*' is the 2D convolution.
#'
#' These filters are decomposible:
#'\deqn{ G_x = t([1 2 1]) * ([-1 0 1] * img) }
#'\deqn{ G_y = t([-1 0 1]) * ([1 2 1] * img) }
#'
#' The sobel image is given by:
#' \deqn{ G = \sqrt{ G_x^2 + G_y^2 } }
#'
#' If the user specifies \code{which='both'} (default), they will get
#' \deqn{G}. If they specify \code{which='horizontal'} they'll get
#' \deqn{G_x}, and if they specify \code{which='vertical'} they'll get
#' \deqn{G_y}.
#'
#' Note: one can also calculate the *direction* of the gradient via:
#' \deqn{ theta = atan2(G_y,G_x) }
imsobel <- function(img, which=c('both','horizontal','vertical'))
{
    which <- match.arg(which)
    # horizontal sobel
    if ( which != 'vertical' ) {
        sobelX <- filter(t(filter(t(img),c(-1,0,1))),c(1,2,1))
    }
    # vertical sobel
    if ( which != 'horizontal' ) {
        sobelY <- filter(t(filter(t(img),c(1,2,1))),c(-1,0,1))
    }
    f <- function(x) array(x,dim=dim(img))
    if ( which=='both' )
        return( f(sqrt( sobelX^2 + sobelY^2 )) )
    else if ( which == 'horizontal' )
        return( f(sobelX) )
    else if ( which == 'vertical' )
        return( f(sobelY) )
}

#' Draw a y=ax+b (origin bottom-left, cartesian) line onto an image
#' 
#' @param img image to draw on
#' @param a slope of line. for the special case \code{a=Inf}, 
#'        the equation of the line is taken to be \code{x=b}.
#' @param b intercept of line
#' @param col colour to draw the line in
#' @param x1 left x clipping box, default left of image
#' @param y1 bottom y clipping box, default bottom of image
#' @param x2 right x clipping box, default right of image
#' @param y2 top y clipping box, default top of image
#' @param one.based whether the input indices are one-based
#' @param max maximum image intensity (can usually ignore)
#' @return the image with the line \code{y=ax+b} drawn on it, within the
#' clipping box (x1,y1) to (x2,y2) (inclusive).
#'
#' @note if \code{col} is supplied in vector form and has any elements
#' greater than 1, *and* max is greater than 1, then we will assume that
#' \code{col} scales from 0 to 255.
#' @family draw
#' @family sandbox
#' @seealso \code{\link{imdrawline}} for a line given specific start/end
#' points
imdrawabline <- function(img, a, b, col,
                         x1=1,y1=imgetny(img),
                         x2=imgetnx(img),y2=1,
                         one.based=TRUE,
                         max=imageUtilities.option('max'))
{
    if ( is.character(col) ) col <- as.vector(col2rgb(col))/255*max
    if ( !is.vector(col) ) col <- as.vector(col)
    if ( any(col>max) ) col <- col/255*max

    Y1 <- min(y1,y2)
    Y2 <- base::max(y1,y2)
    X1 <- min(x1,x2)
    X2 <- base::max(x1,x2)
    if ( !one.based ) {
        X1 <- X1+1
        Y1 <- Y1+1
        X2 <- X2+1
        Y2 <- Y2+1
    }

    imgOut <- imexpand(img)

    # flip upside down for now (because of top-left vs bottom-left
    # coordinate system)
    imgOut <- imgOut[imgetny(img):1,,]

    # Special cases - horizontal or vertical

    # *** @BIGTODO CHECK 1-BASED
    # y=x should start from 1,1   2,2    3,3
    # y=3 should be *FROM TOP OR FROM BOTTOM*?
    # horizontal. y=1 will be pixel 1.
    if ( all.equal(a,0) ) {
        imgOut[b,Y1:Y2,] <- col
    } else if ( all.equal(a,Inf) ) {
    # vertical
    # x=1 is first clumn
        imgOut[X1:X2,b,] <- col
    } else {
        # a=1: x=x1:x2
        # a=2: x1:.5:x2 ?
        x <- seq(X1,X2,by=1/a)
        y <- round(a*x+b)
        # clip to boundin box
        which.keep <- (y>=Y1 & y<=Y2) 
        x <- round(x[which.keep])
        y <- y[which.keep]

        # draw.
        # index = sub2ind(dim(img),round(y),round(x))
        idx <- matrix(c(x,y),ncol=2) # by column.
        imgOut[idx] <- col
    }
    # flip image back
    imgOut <- imgOut[imgetny(img):1,,]
    return(imsqueeze(imgOut))
}

#' Draw a line from point a to point b on an image
#' @param img image to draw on
#' @param x1 column to start the line
#' @param y1 row to start the line
#' @param x2 column to end the line
#' @param y2 row to end the line
#' @param col colour to draw with
#' @param one.based boolean, whether x1, x2, y1, and y2 are 1-based
#' @param max the maximum intensity of the image
#' @return the image with the line from x1,x2 to y1,y2 on it (inclusive).
#'
#' @note if \code{col} is supplied in vector form and has any elements
#' greater than 1, *and* max is greater than 1, then we will assume that
#' \code{col} scales from 0 to 255.
#' @family draw
#' @family sandbox
#' @seealso \code{\link{imdrawabline}} for a y=ax+b image
imdrawline <- function(img, x1, y1, x2, y2, col,
                       one.based=TRUE,
                       max=imageUtilities.option('max'))
{
    # CHECK FOR HORIZ/VERT
    # work out the y=ax+b
    a <- (y2-y1)/(x2-x1)
    b <- y1-a*x1

    return(imdrawabline(img, a, b, col,
                        x1=x1, y1=y1, x2=x2, y2=y2,
                        one.based=one.based,
                        max=max))

}


#' plot points onto an image with lines around them (NOT YET WORKING)
#' 
#' @param img    the image to draw on to.
#' @param ipts   an nx2 matrix with points to be drawn on the image. Column
#'               names "X" and "Y".
#' @param radius the radius of points to draw (pixels). CURRENTLY NOT
#' RESPECTED.
#' @param lines  boolean,whether to draw lines in between the points
#' @param transparency The transparency of the points. 0 for opaque, 1 for
#'         transparent.
#' @param col     The colour to draw the overlay in.
#'                Either a string of a colour recognised by R, or
#'                 a numeric vector \code{c(R,G,B)} with red/green/blue
#'                 values the desired colour, from 0 to 255.
#' @param one.based whether the input points are one-based or not.
#' @return image with the lines and points drawn on.
#' @family draw
plotPoints <- function( img, ipts,
                       radius=min(imgetnx(img,1),imgetny(img,1))/100,
                       lines=FALSE, transparency=0, col='red',
                       one.based=TRUE ) 
{
    # USE GRID.GRAPHICS
    # 1) can draw with grid, save plot, and load back again to get pixels
    # 2) since it's just lines and points --- do it yourself.
    msk <- array(0,dim=dim(img)[1:2])
    pts <- ipts
    if ( nrow(ipts)==2 && ncol(ipts)!=2 ) 
        pts <- t(ipts)
    if ( !is.null(colnames(pts)) )
        pts <- pts[,c("X","Y")]
    else
        colnames(pts) <- c("X","Y")

    pts <- sortPointsRectangle( pts )[,c("X","Y")]

    if ( lines ) {
    # @TODO
    #imdrawcircle( img, pts )
        for ( i in 1:nrow(pts) ) {
            j <- ifelse( i==nrow(pts), 1, i+1 )
            img <- imdrawline(img, pts[i,"X"], pts[i,"Y"],
                                   pts[j,"X"], pts[j,"Y"],
                                   col,
                                   one.based=one.based)
        }
    }
    # TODO:
    # DILATE ?!?!?!?!
    #msk <- imbinary(msk)
    #if ( radius>1 ) 
    #    msk <- !imbderode(!msk,radius)
  #      msk <- imbddilate( msk, radius )

    return(img)
}
